#######################################################################
###### Plugin Software 'FJpipeline' was Designed and Developed by
###### BDBI, Product Innovation.
###### Authored by Joshua Luthy
###### This template was last updated in August 2022
#######################################################################
runID <- "FJ_RUNID"
#######################################################################
###### Create the FlowJo Object
#######################################################################
setClass("FJObj",
         slots = c(
           runID = "character", # Tie this into my Noir package to use this object with FlowJo server.
           # pgModularity = "integer",
           inputCSV = "list",  # create some slots to hold data for later use
           flowFrameFJ = "list",
           BatchCorrect = "list",
           PeacoQCPops = "list",
           FlowAIPops = "list",
           Rphenograph = "list",
           FlowSOMClusters = "list",
           EmbedSOMReduction = "list",
           ClustRCheck = "list",
           Taylor = "list"
         ),
         prototype = list(
           runID = "runID",
           inputCSV = list(X=matrix(),
                           cellIds=matrix(),
                           eventNumbers=matrix()
           ),
           flowFrameFJ = list(ffFJ=list()
           ),
           BatchCorrect = list(corrected.asinh.transformed=matrix(),
                               corrected.reverse.transformed=matrix(),
                               markersToNorm="character",
                               labels=matrix(),
                               id=matrix(),
                               colNames=matrix()

           ),
           PeacoQCPops = list(results=matrix()
           ),
           FlowAIPops = list(goodEvents=matrix(),
                             badEvents=matrix()
           ),
           Rphenograph = list(clusters=matrix(),
                              clustNumber=matrix(),
                              modularity=matrix()
           ),
           FlowSOMClusters = list(clusters=matrix(),
                                  clustNumber=matrix(),
                                  som=list()
           ),
           EmbedSOMReduction = list(embeddingMat=matrix(),
                                    embeddingType="character"
           ),
           ClustRCheck = list(values=matrix()
           ),
           Taylor = list(values=matrix()
           )
         ),
)
#######################################################################
##         LOAD INPUTS  & FUNCTIONS     ##
#######################################################################
userInputs <- function(object){
  cat("RunID:", object@runID, "\n")
  cat(object@EmbedSOMReduction$embeddingType, "\n")
}

setGeneric("userInputs", function(object) {
  standardGeneric("userInputs")
})

setMethod("userInputs",
          "FJObj",
          function(object){
            cat("RunID:", object@runID, "\n")
            cat(object@EmbedSOMReduction$embeddingType, "\n")
          }
)

setGeneric("loadFlowJoCSV", function(object, csvPath) {
  standardGeneric("loadFlowJoCSV")
})

setMethod("loadFlowJoCSV",
          "FJObj",
          function(object, csvPath){
            data <- read.csv(csvPath, check.names=FALSE)
            cellIdsColumn <- rep(NA, nrow(data))
            cellIds <- data.frame(CellIdDP = cellIdsColumn)
            names(cellIds) <- CellIdDP
            if (CellIdDP %in% names(data)) {
              cellIds <- data[, CellIdDP]
              data <- data[, -which(names(data) %in% c(CellIdDP))]
            }

            EventNumberColumn <- rep(NA, nrow(data))
            eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
            names(eventNumbers) <- EventNumberDP
            if (EventNumberDP %in% names(data)) {
              eventNumbers <- data[, EventNumberDP]
              data <- data[, -which(names(data) %in% c(EventNumberDP))]
            }

            # Trim the leading and tailing whitespaces
            colnames(data) <- gsub("^\\s+|\\s+$", "", colnames(data))

            object@inputCSV <- list(X=as.matrix(data),cellIds=as.matrix(cellIds),eventNumbers=as.matrix(eventNumbers))
            colnames(object@inputCSV$cellIds) <- "CellId"
            colnames(object@inputCSV$eventNumbers) <- "EventNumberDP"
            return(object)
          }

)

setGeneric("loadFlowJoList", function(object, listElement) {
  standardGeneric("loadFlowJoList")
})

setMethod("loadFlowJoList",
          "FJObj",
          function(object, listElement){
            data <- listElement
            cellIdsColumn <- rep(NA, nrow(data))
            cellIds <- data.frame(CellIdDP = cellIdsColumn)
            names(cellIds) <- CellIdDP
            if (CellIdDP %in% names(data)) {
              cellIds <- data[, CellIdDP]
              data <- data[, -which(names(data) %in% c(CellIdDP))]
            }

            EventNumberColumn <- rep(NA, nrow(data))
            eventNumbers <- data.frame(EventNumberDP = EventNumberColumn)
            names(eventNumbers) <- EventNumberDP
            if (EventNumberDP %in% names(data)) {
              eventNumbers <- data[, EventNumberDP]
              data <- data[, -which(names(data) %in% c(EventNumberDP))]
            }

            # Trim the leading and tailing whitespaces
            # colnames(data) <- gsub("^\\s+|\\s+$", "", colnames(data))

            object@inputCSV <- list(X=as.matrix(data),cellIds=as.matrix(cellIds),eventNumbers=as.matrix(eventNumbers))
            colnames(object@inputCSV$cellIds) <- "CellId"
            colnames(object@inputCSV$eventNumbers) <- "EventNumberDP"
            return(object)
          }

)

setGeneric("flowFrameFromFlowJoCSV", function(object, filename, dropCompPrefix=FALSE){
  standardGeneric("flowFrameFromFlowJoCSV")
})
################################################################
#    Build a FlowFrame  #
################################################################
#TODO make generateGUID from FlowSOM method work below.
## Generates GUID like "47689979-e471-47a6-aa1a-7910d2d42b03"
## As a side effect, this also sets the seed based in the data passed in
generateGUID <- function(data) {
  mySeed <- round(1000*mean(as.matrix(data))) + dim(data)[1] + dim(data)[2]
  while (mySeed > .Machine$integer.max) mySeed <- round(mySeed/3)
  set.seed(as.integer(mySeed))

  part1 <- paste(unlist(lapply(1:8, function(x) { sprintf("%x", round(runif(1, 0, 9))) })), collapse="", sep="")
  part2 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  part3 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  part4 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  part5 <- paste(unlist(lapply(1:12, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  paste(part1, part2, part3, part4, part5, sep="-")
}

setMethod("flowFrameFromFlowJoCSV",
          "FJObj",
          function(object, filename, dropCompPrefix=FALSE){
            data <- object@inputCSV$X
            colNames <- colnames(data)
            npar <- length(colNames)
            id <- paste("$P",1:npar,sep="")

            maxRange <- unlist(lapply(1:npar, function(x) {
              max(data[,x])
            }))
            range <- maxRange + 1
            minRange <- unlist(lapply(1:npar, function(x) {
              min(0, min(data[,x]))
            }))

            colNames <- gsub("^\\s+|\\s+$", "", colNames) 	# Trim the leading and tailing whitespaces

            ## FlowJo splits parameter names and description by " :: ", e.g., "Comp-PE-Cy5-A :: CD19#PE-Cy5"
            names <- unlist(lapply(1:npar, function(i) {
              unlist(strsplit(colNames[i], " :: "))[1]
            }))
            ## FlowJo indicates that a patameter is compensated by the Comp- prefix in its name, we can drop that if undesirable
            if (dropCompPrefix) {
              names <- gsub("^\\Comp-", "FJNorm-", names)
            }
            ## Again splitting by " :: " but this time taking the second part (if present) as description
            desc <- unlist(lapply(1:npar, function(i) {
              s <- unlist(strsplit(colNames[i], " :: "))
              if (length(s) >= 2) s[2]
              else NA
            }))

            pars <- new("AnnotatedDataFrame",
                        data=data.frame(
                          row.names=I(id), name=I(names), desc=I(desc),
                          range=range, minRange=minRange, maxRange=maxRange),
                        varMetadata=data.frame(
                          row.names=I(c("name","desc","range", "minRange", "maxRange")),
                          labelDescription=I(c("Name of Parameter", "Description of Parameter", "Range of Parameter", "Minimum Parameter Value after Transforamtion", "Maximum Parameter Value after Transformation"))
                        )
            )

            ## Build the descripton
            txt <- list()
            txt["FCSversion"] <- "3"
            txt["$PAR"] <- as.character(npar)
            txt["$TOT"] <- as.character(dim(data)[1])
            txt["$BYTEORD"] <- "4,3,2,1"
            txt["$DATATYPE"] <- "F"
            txt["$MODE"] <- "L"
            txt["$NEXTDATA"] <- "0"

            txt["$BEGINANALYSIS"] <- "0"
            txt["$ENDANALYSIS"] <- "0"
            txt["$BEGINSTEXT"] <- "0"
            txt["$ENDSTEXT"] <- "0"
            dataStart <- 1000 ## This doesn't really matter, it's not an FCS file anyway, but let's provide something reasonable
            txt["$BEGINDATA"] <- as.character(dataStart)
            txt["$ENDDATA"] <- as.character(dataStart + npar * dim(data)[1] * 4 - 1) ## $DATATYPE is F and PnB is 32, so here we do: start + rows x cols x 4 bytes per value - 1

            txt[paste(id,"N",sep="")] <- names
            txt[paste(id,"S",sep="")] <- desc
            txt[paste(id,"R",sep="")] <- as.character(range)
            txt[paste(id,"B",sep="")] <- "32"
            txt[paste(id,"E",sep="")] <- "0,0"
            txt[paste(id,"G",sep="")] <- "1"
            txt[paste("flowCore_",id,"Rmin",sep="")] <- as.character(minRange)
            txt[paste("flowCore_",id,"Rmax",sep="")] <- as.character(maxRange)

            txt["$FIL"] <- basename(filename)
            txt["FILENAME"] <- object@runID
            txt["$SYS"] <- paste(Sys.info()['sysname'], Sys.info()['release'])
            txt["CREATOR"] <- R.version.string
            txt["transformation"] <- "applied"
            txt["GUID"] <- generateGUID(data)
            txt["ORIGINALGUID"] <- generateGUID(data)
            txt["FJRUNID"] <- object@runID

            txt["$INST"] <- "Unknown institution"
            txt["$OP"] <- "Unknown operator"
            txt["$CYT"] <- "Unknown cytometer"
            txt["$SRC"] <- "Unknown source"

            curtime <- Sys.time()
            txt["$DATE"] <- toupper(format(curtime, "%d-%b-%Y")) ## Current data in the format of "05-OCT-2017"
            timeDelta <- 120 ## picking 2 minutes as cytometer run time
            if ("time" %in% tolower(names)) {
              txt["$TIMESTEP"] <- "0.01" ## We don't know that, so just something that would be reasonable-ish
              timeIndex <- which(tolower(names) == "time")
              timeDelta <- ceiling((max(data[,timeIndex]) - min(data[,timeIndex])) * 0.01)
            }
            if (toupper(format(curtime - timeDelta, "%d-%b-%Y")) == toupper(format(curtime, "%d-%b-%Y"))) {
              ## Subtracting timeDelta is still the same day, so let's do it that way.
              txt["$BTIM"] <- format(curtime - timeDelta, "%H:%M:%S")
              txt["$ETIM"] <- format(curtime, "%H:%M:%S")
            } else {
              ## Subtracting timeDelta gets us to yesterday, that's not good, so let's just say we started at midnight and add delta to it as ETIM...
              ## (all this is arbitrary anyway, we just want resonable values)
              txt["$BTIM"] <- "00:00:00"
              txt["$ETIM"] <- sprintf("%02i:%02i:%02i", floor(timeDelta / 3600), floor((timeDelta%%3600) / 60), floor(timeDelta%%60))
            }
            colnames(data) <- names
            object@flowFrameFJ <- list(fcs=new("flowFrame",
                                               exprs=as.matrix(data),
                                               parameters=pars,
                                               description=txt),
                                       cellIds=object@inputCSV$eventNumbers)

            return(object)
          })


################################################################
#    Perform Batch Correction  #
################################################################

# setGeneric("runBatch_correct", function(object){
#   standardGeneric("runBatch_correct")
# })
#
# setMethod("runBatch_correct",
#           "FJObj",
#           function(object, fjBatchID, fjCofactor, outputDir){
#             exprssnMatrixUncorrected <- as.data.frame(object@flowFrameFJ[["fcs"]]@exprs)
#             markersToNorm <- colnames(exprssnMatrixUncorrected[,1:fjBatchID-1])
#             object@BatchCorrect$markersToNorm <- markersToNorm
#
#             uncorrected <- transform_asinh(
#               df = as.data.frame(object@flowFrameFJ[["fcs"]]@exprs),
#               markers = myFJobj@BatchCorrect$markersToNorm,
#               cofactor = fjCofactor,
#               .keep = TRUE # Lets you keep all columns, in case they are useful to you
#             )
#             uncorrected$batch <- as.factor(uncorrected[,fjBatchID])
#             # Run batch correction
#             object@BatchCorrect$labels <- uncorrected %>%
#               normalize(markers = object@BatchCorrect$markersToNorm,
#                         norm_method = "scale") %>%
#               create_som(markers = object@BatchCorrect$markersToNorm,
#                          rlen = 10)
#
#             object@BatchCorrect$corrected.asinh.transformed <- uncorrected %>%
#               correct_data(label = object@BatchCorrect$labels)
#
#             plot_density(uncorrected = uncorrected,
#                          corrected = corrected,
#                          markers = markersToNorm,
#                          filename = paste0(outputDir, "Density_plot_",fjCofactor,"_.png" ))
#
#             return(object)
#
#           }
# )


