#######################################################################
###### Plugin Software was Designed and Developed by
###### BDBI, Product Innovation.
###### Authored by Joshua Luthy
###### This template was last updated in March 2023
#######################################################################
##         LOAD INPUTS  & FUNCTIONS     ##
#######################################################################
# setwd("FJ_OUTPUT_DIRECTORY_PATH")
outPutFolder <- "FJ_OUTPUT_DIRECTORY_PATH"
inputCSVpath <- "FJ_SAMPLE_DATA_PATH"
runID <- "FJ_RUNID"
fjPopName <- "FJ_POPNAME"
# Batch Correction
runBatchCorrection <- TRUE
fMarkersToNorm <- c(FJ_MARKER_LIST)
fBatchID <- "FJ_BATCH_ID"
fjcofactor <- FJ_COFACTOR
fRunPerformanceChecks <- FJ_P_METRIX
fjXdim <- FJ_XDIM
fjYdim <- FJ_YDIM
fjNormMethod <- "FJ_NORM_METHOD"

# Constants
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellId"

source("FJ_METHODSFILE_PATH")

StopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(outPutFolder, "Error_Message_from_R.txt"))
  stop(message, ...)
}

ReturnErrorMessage <- function(message, ...) {
  cat(message, file=paste0(outPutFolder, "Plotting_Error_Message_from_R.txt"))
}

normMethodUsed <- function(FJinput){
  switch(FJinput,
         scale={
           nm <- "Z-score"
         },
         rank={
           nm <- "Rank"
         })
  return(nm)
}
#######################################################################
###### Start package import or installations
#######################################################################
# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need, repos="https://cran.r-project.org")
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}
tryCatch({
  loadPackages(c('dplyr','readr','BiocManager','tibble','cowsay','data.table','cowplot','devtools','tidyr','ggridges', 'emdist'))
}, error = function(e) {
  message <- "There was an error with installing R packages. Please see How-To PDF and try to manually install."
  StopAndReturnError(message = message)
}, warning = function(w){
  loadPackages(c('dplyr','readr','BiocManager','tibble','cowsay','data.table','cowplot','devtools','tidyr', 'ggridges', 'emdist'))
  writeLines("loaded R packages")
})

tryCatch({
  loadBioc(c('flowCore', 'sva'))
}, error = function(e) {
  message <- "There was an error with installing R packages. Please see How-To PDF and try to manually install."
  StopAndReturnError(message = message)
}, warning = function(w) {
  loadBioc(c('flowCore', 'sva')) # , 'FlowSOM', 'ComplexHeatmap', 'PeacoQC'
  writeLines("loaded R packages")
})
# The below should work for R-4.1 and later.
if(!require("cyCombine",character.only=TRUE)){
  suppressMessages(suppressWarnings(devtools::install_github('biosurf/cyCombine')))
}

tryCatch({
  library("cyCombine")
}, error = function(e) {
  message <- "CyCombine was not able to run. There is an error loading the cyCombine R package. Please see How-To PDF and try to manually install or contact flowjo@bd.com."
  StopAndReturnError(message = message)
})

ev <- as.character(packageVersion("cyCombine"))
if(compareVersion(ev, "0.2.13") < 0 || compareVersion(ev, "3") >=0)
  StopAndReturnError("Incompatible cyCombine version. Must be v0.2.13 or later.")

sessionInfo()

#######################################################################
###### Start the File Import
#######################################################################
myFJobj <- new("FJObj", runID = runID)  # Create the basic FlowJo object ad fill slots as we go
myFJobj <- loadFlowJoCSV(myFJobj, inputCSVpath)

###############################################################
#  Batch Correction with CyCombine #
###############################################################
if(runBatchCorrection){
  # get colnames for sorting later
  myFJobj@BatchCorrect$colNames <- colnames(myFJobj@inputCSV$X)
  # store the marker names to normalize
  myFJobj@BatchCorrect$markersToNorm <- fMarkersToNorm
  # Transform data - This function also de-randomizes the data
  uncorrected <- transform_asinh(
    df = as.data.frame(myFJobj@inputCSV$X),
    markers = myFJobj@BatchCorrect$markersToNorm,
    cofactor = fjcofactor,
    .keep = TRUE # Lets you keep all columns, in case they are useful to you
  )
  # create BatchID column to factor needed for CyCombine
  uncorrected$batch <- as.factor(uncorrected[,fBatchID])
  # Run batch correction
  myFJobj@BatchCorrect$labels <- uncorrected %>%
    normalize(markers = myFJobj@BatchCorrect$markersToNorm,
              norm_method = "FJ_NORM_METHOD") %>%
    create_som(markers = myFJobj@BatchCorrect$markersToNorm,
               rlen = 10,
               xdim = fjXdim,
               ydim = fjYdim)

  # 'covar' is the optional second keyword present in the exported data.
  # Here we do not have a covar keyword present so it is removed.
  myFJobj@BatchCorrect$corrected.asinh.transformed <- uncorrected %>%
          correct_data(label = myFJobj@BatchCorrect$labels,
                       markers = myFJobj@BatchCorrect$markersToNorm)#,
  # covar = "status")

  uncorrected$label <- myFJobj@BatchCorrect$labels
  uncorrected$id <- myFJobj@BatchCorrect$corrected.asinh.transformed$id
  # Plots can be made like so
    tryCatch(
    {

        png("NUL")
        pdf("NUL")
    },
    error=function(cond) {

    },
    finally={
    }
    )
  plot_density(uncorrected = uncorrected,
               corrected = myFJobj@BatchCorrect$corrected.asinh.transformed,
               markers = myFJobj@BatchCorrect$markersToNorm,
               filename = paste0(outPutFolder, "Density_plot_", fjPopName, "_cofactor_" , fjcofactor,"_.png" ))

  tryCatch({
    plot_density(uncorrected = uncorrected,
                 corrected = myFJobj@BatchCorrect$corrected.asinh.transformed,
                 markers = myFJobj@BatchCorrect$markersToNorm,
                 filename = paste0(outPutFolder, "Density_plot_", fjPopName, "_cofactor_" , fjcofactor,"_.png" ))
  }, error = function(e) {
    message <- "There was an error with generation of png file."
    ReturnErrorMessage(message)
  })


  # Transform back to scaled values before export
  myFJobj@inputCSV$X <- transform_asinh(
    df = myFJobj@BatchCorrect$corrected.asinh.transformed,
    markers = myFJobj@BatchCorrect$markersToNorm,
    cofactor = fjcofactor,
    reverse = TRUE,
    .keep = TRUE # Lets you keep all columns, in case they are useful to you
  )

  if(fRunPerformanceChecks){
    # Evaluate results with EMD and MAD scores
    emdCalc <- NULL
    if(nrow(myFJobj@inputCSV$X) > 100000){
      emdCalc <- NULL
      message <- "Too many events OR not enough memory to calculate earth movers distances(EMD)."
      ReturnErrorMessage(message)
    } else {
      # Evaluate EMD
      tryCatch({
        # png("NUL")
        # pdf("NUL")
        emdCalc <- evaluate_emd(uncorrected, myFJobj@BatchCorrect$corrected.asinh.transformed, cell_col = "label")
      }, error = function(e) {
        emdCalc <- NULL
        message <- "EMD calculation failed. Possibly too many events OR not enough memory to calculate earth movers distances(EMD)."
        ReturnErrorMessage(message)
      })
    }

    if(is.null(emdCalc)) {
      emdScore <- NULL
      violinPlot <- NULL
      scatters <- NULL
    } else {
      # Reduction
      emdScore <- emd$reduction
      # Violin plot
      violinPlot <- emd$violin
      # Scatter plot
      scatters <- emd$scatter
      ggplot2::ggsave(file=paste0(outPutFolder, "ViolinQCPlot_", fjPopName, "_.png"), device= "png", plot= violinPlot, dpi=300, bg = "transparent")
      ggplot2::ggsave(file=paste0(outPutFolder, "ScatterQCPlot_", fjPopName, "_.png"), device= "png", plot= scatters, dpi=300, bg = "transparent")
    }
    # Evaluate MAD
    if(nrow(myFJobj@inputCSV$X) > 100000){
      mad <- NULL
      message <- "Too many events or not enough memory to calculate MAD score."
      ReturnErrorMessage(message)
    } else {
      tryCatch({
        mad <- evaluate_mad(uncorrected, myFJobj@BatchCorrect$corrected.asinh.transformed, cell_col = "label")
      }, error = function(e) {
        mad <- NULL
        message <- "There was an error with MAD calculation, try running on smaller population."
        ReturnErrorMessage(message)
      })
    }

    if(is.null(mad)){
      madScore <- NULL
    } else {
      # Score
      madScore <- mad$score
    }

    if(is.null(emdCalc) && is.null(mad)){
      message <- c("Run Settings\n",
                   "Population Selected:", fjPopName, "\n",
                   "Batch Identifier:", fBatchID, "\n",
                   "Cofactor for Asinh Transform:", fjcofactor, "\n",
                   "SOM X Grid Size:", fjXdim, "\n",
                   "SOM Y Grid Size:", fjYdim, "\n",
                   "Normalization Method:", normMethodUsed(fjNormMethod), "\n",
                   "EMD Score: failed", "\n",
                   "MAD Score: failed")
      cat(message, file=paste0(outPutFolder, "RunSettings_", runID, "_.txt"))
    } else if (is.null(emdCalc) && !is.null(mad)){
      message <- c("Run Settings\n",
                   "Population Selected:", fjPopName, "\n",
                   "Batch Identifier:", fBatchID, "\n",
                   "Cofactor for Asinh Transform:", fjcofactor, "\n",
                   "SOM X Grid Size:", fjXdim, "\n",
                   "SOM Y Grid Size:", fjYdim, "\n",
                   "Normalization Method:", normMethodUsed(fjNormMethod), "\n",
                   "EMD Score: failed", "\n",
                   "MAD Score:", madScore)
      cat(message, file=paste0(outPutFolder, "RunSettings_", runID, "_.txt"))
    } else if (!is.null(emdCalc) && is.null(mad)){
      message <- c("Run Settings\n",
                   "Population Selected:", fjPopName, "\n",
                   "Batch Identifier:", fBatchID, "\n",
                   "Cofactor for Asinh Transform:", fjcofactor, "\n",
                   "SOM X Grid Size:", fjXdim, "\n",
                   "SOM Y Grid Size:", fjYdim, "\n",
                   "Normalization Method:", normMethodUsed(fjNormMethod), "\n",
                   "EMD Score:", emdScore, "\n",
                   "MAD Score: failed")
      cat(message, file=paste0(outPutFolder, "RunSettings_", runID, "_.txt"))
    } else {
      message <- c("Run Settings\n",
                   "Population Selected:", fjPopName, "\n",
                   "Batch Identifier:", fBatchID, "\n",
                   "Cofactor for Asinh Transform:", fjcofactor, "\n",
                   "SOM X Grid Size:", fjXdim, "\n",
                   "SOM Y Grid Size:", fjYdim, "\n",
                   "Normalization Method:", normMethodUsed(fjNormMethod), "\n",
                   "EMD Score:", emdScore, "\n",
                   "MAD Score:", madScore)
      cat(message, file=paste0(outPutFolder, "RunSettings_", runID, "_.txt"))
    }
  } else {
    message <- c("Run Settings\n",
                 "Population Selected:", fjPopName, "\n",
                 "Batch Identifier:", fBatchID, "\n",
                 "Cofactor for Asinh Transform:", fjcofactor, "\n",
                 "SOM X Grid Size:", fjXdim, "\n",
                 "SOM Y Grid Size:", fjYdim, "\n",
                 "Normalization Method:", normMethodUsed(fjNormMethod), "\n",
                 "EMD Score: Not Calculated", "\n",
                 "MAD Score: Not Calculated")

    cat(message, file=paste0(outPutFolder, "RunSettings_", runID, "_.txt"))
  }
  # Remove columns added by CyCombine before
  myFJobj@inputCSV$X[["id"]] <- NULL
  myFJobj@inputCSV$X[["label"]] <- NULL
  myFJobj@inputCSV$X[["batch"]] <- NULL

  # get the original ordering before exporting
  paramOrderIdx <- match(myFJobj@BatchCorrect$colNames, colnames(myFJobj@inputCSV$X))
  # Replace original expression matrix with batch corrected one
  myFJobj@inputCSV$X <- myFJobj@inputCSV$X[ ,paramOrderIdx]
  # create a flowFrame
  myFJobj <- flowFrameFromFlowJoCSV(myFJobj, inputCSVpath, dropCompPrefix=FALSE)

  # Now export as new .fcs file.
  write.FCS(x =  myFJobj@flowFrameFJ[["fcs"]],
            filename = paste0(outPutFolder, "Normalized_", fjPopName,"_", runID, "_", ".fcs"),
            what="numeric",
            delimiter = "|",
            endian="big")

  # Clean up to make the object light
  myFJobj@BatchCorrect$corrected.asinh.transformed <- as.matrix(0)
  myFJobj@BatchCorrect$corrected.reverse.transformed <- as.matrix(0)
  myFJobj@inputCSV$X <- as.matrix(0)

}

###############################################################
#  Exporting of results to a csv file to be merged to FJ   #
###############################################################
saveRDS(myFJobj, file = paste0(outPutFolder, "FlowJo_Object_",runID,"_",fjPopName,".RDS"), compress = TRUE)
say("time")